package sun.reflect.annotation;

import java.nio.BufferUnderflowException;
import java.lang.annotation.AnnotationFormatError;
import java.lang.annotation.RetentionPolicy;
import java.util.LinkedHashMap;
import java.util.Map;
import java.nio.ByteBuffer;
import sun.misc.JavaLangAccess;
import sun.misc.SharedSecrets;
import java.util.Iterator;
import java.util.List;
import java.lang.reflect.Executable;
import java.lang.annotation.Annotation;
import java.lang.reflect.GenericDeclaration;
import java.util.Arrays;
import java.util.ArrayList;
import java.lang.reflect.AnnotatedType;
import java.lang.reflect.Type;
import java.lang.reflect.AnnotatedElement;
import sun.reflect.ConstantPool;

public final class TypeAnnotationParser
{
    private static final TypeAnnotation[] EMPTY_TYPE_ANNOTATION_ARRAY;
    private static final byte CLASS_TYPE_PARAMETER = 0;
    private static final byte METHOD_TYPE_PARAMETER = 1;
    private static final byte CLASS_EXTENDS = 16;
    private static final byte CLASS_TYPE_PARAMETER_BOUND = 17;
    private static final byte METHOD_TYPE_PARAMETER_BOUND = 18;
    private static final byte FIELD = 19;
    private static final byte METHOD_RETURN = 20;
    private static final byte METHOD_RECEIVER = 21;
    private static final byte METHOD_FORMAL_PARAMETER = 22;
    private static final byte THROWS = 23;
    private static final byte LOCAL_VARIABLE = 64;
    private static final byte RESOURCE_VARIABLE = 65;
    private static final byte EXCEPTION_PARAMETER = 66;
    private static final byte INSTANCEOF = 67;
    private static final byte NEW = 68;
    private static final byte CONSTRUCTOR_REFERENCE = 69;
    private static final byte METHOD_REFERENCE = 70;
    private static final byte CAST = 71;
    private static final byte CONSTRUCTOR_INVOCATION_TYPE_ARGUMENT = 72;
    private static final byte METHOD_INVOCATION_TYPE_ARGUMENT = 73;
    private static final byte CONSTRUCTOR_REFERENCE_TYPE_ARGUMENT = 74;
    private static final byte METHOD_REFERENCE_TYPE_ARGUMENT = 75;
    
    public TypeAnnotationParser() {
    }
    
    public static AnnotatedType buildAnnotatedType(byte[] array, ConstantPool constantPool, AnnotatedElement annotatedElement, Class<?> clazz, Type type, TypeAnnotation.TypeAnnotationTarget typeAnnotationTarget) {
        TypeAnnotation[] typeAnnotations = parseTypeAnnotations(array, constantPool, annotatedElement, clazz);
        ArrayList list = new ArrayList(typeAnnotations.length);
        for (TypeAnnotation typeAnnotation : typeAnnotations) {
            if (typeAnnotation.getTargetInfo().getTarget() == typeAnnotationTarget) {
                list.add(typeAnnotation);
            }
        }
        TypeAnnotation[] array2 = (TypeAnnotation[])list.toArray(TypeAnnotationParser.EMPTY_TYPE_ANNOTATION_ARRAY);
        return AnnotatedTypeFactory.buildAnnotatedType(type, TypeAnnotation.LocationInfo.BASE_LOCATION, array2, array2, annotatedElement);
    }
    
    public static AnnotatedType[] buildAnnotatedTypes(byte[] array, ConstantPool constantPool, AnnotatedElement annotatedElement, Class<?> clazz, Type[] array2, TypeAnnotation.TypeAnnotationTarget typeAnnotationTarget) {
        int length = array2.length;
        AnnotatedType[] a = new AnnotatedType[length];
        Arrays.fill(a, AnnotatedTypeFactory.EMPTY_ANNOTATED_TYPE);
        ArrayList[] array3 = new ArrayList[length];
        TypeAnnotation[] typeAnnotations;
        for (TypeAnnotation e : typeAnnotations = parseTypeAnnotations(array, constantPool, annotatedElement, clazz)) {
            TypeAnnotation.TypeAnnotationTargetInfo targetInfo = e.getTargetInfo();
            if (targetInfo.getTarget() == typeAnnotationTarget) {
                int count = targetInfo.getCount();
                if (array3[count] == null) {
                    array3[count] = new ArrayList(typeAnnotations.length);
                }
                array3[count].add(e);
            }
        }
        TypeAnnotation[] empty_TYPE_ANNOTATION_ARRAY;
        for (int j = 0; j < length; a[j] = AnnotatedTypeFactory.buildAnnotatedType(array2[j], TypeAnnotation.LocationInfo.BASE_LOCATION, empty_TYPE_ANNOTATION_ARRAY, empty_TYPE_ANNOTATION_ARRAY, annotatedElement), ++j) {
            ArrayList<Object> list = array3[j];
            if (list != null) {
                empty_TYPE_ANNOTATION_ARRAY = (TypeAnnotation[])list.toArray(new TypeAnnotation[list.size()]);
            }
            else {
                empty_TYPE_ANNOTATION_ARRAY = TypeAnnotationParser.EMPTY_TYPE_ANNOTATION_ARRAY;
            }
        }
        return a;
    }
    
    public static AnnotatedType buildAnnotatedSuperclass(byte[] array, ConstantPool constantPool, Class<?> clazz) {
        Type genericSuperclass = clazz.getGenericSuperclass();
        if (genericSuperclass == null) {
            return AnnotatedTypeFactory.EMPTY_ANNOTATED_TYPE;
        }
        return buildAnnotatedType(array, constantPool, clazz, clazz, genericSuperclass, TypeAnnotation.TypeAnnotationTarget.CLASS_EXTENDS);
    }
    
    public static AnnotatedType[] buildAnnotatedInterfaces(byte[] array, ConstantPool constantPool, Class<?> clazz) {
        if (clazz == Object.class || clazz.isArray() || clazz.isPrimitive() || clazz == Void.TYPE) {
            return AnnotatedTypeFactory.EMPTY_ANNOTATED_TYPE_ARRAY;
        }
        return buildAnnotatedTypes(array, constantPool, clazz, clazz, clazz.getGenericInterfaces(), TypeAnnotation.TypeAnnotationTarget.CLASS_IMPLEMENTS);
    }
    
    public static <D extends GenericDeclaration> Annotation[] parseTypeVariableAnnotations(D obj, int n) {
        Class<Object> clazz;
        TypeAnnotation.TypeAnnotationTarget typeAnnotationTarget;
        if ((obj instanceof Class)) {
            clazz = (D)obj;
            typeAnnotationTarget = TypeAnnotation.TypeAnnotationTarget.CLASS_TYPE_PARAMETER;
        }
        else if ((obj instanceof Executable)) {
            clazz = (D)obj;
            typeAnnotationTarget = TypeAnnotation.TypeAnnotationTarget.METHOD_TYPE_PARAMETER;
        }
        else {
            throw new AssertionError("Unknown GenericDeclaration " + obj + "\nthis should not happen.");
        }
        List<TypeAnnotation> filter = TypeAnnotation.filter(parseAllTypeAnnotations(clazz), typeAnnotationTarget);
        ArrayList o = new ArrayList(filter.size());
        for (TypeAnnotation typeAnnotation : filter) {
            if (typeAnnotation.getTargetInfo().getCount() == n) {
                o.add(typeAnnotation.getAnnotation());
            }
            else {
                continue;
            }
        }
        return (Annotation[])o.toArray(new Annotation[0]);
    }
    
    public static <D extends GenericDeclaration> AnnotatedType[] parseAnnotatedBounds(Type[] array, D n, int n2) {
        return parseAnnotatedBounds(array, n, n2, TypeAnnotation.LocationInfo.BASE_LOCATION);
    }
    
    private static <D extends GenericDeclaration> AnnotatedType[] parseAnnotatedBounds(Type[] array, D n, int n2, TypeAnnotation.LocationInfo locationInfo) {
        List<TypeAnnotation> fetchBounds = fetchBounds(n);
        if (array != null) {
            int n3 = 0;
            AnnotatedType[] array2 = new AnnotatedType[array.length];
            if (array.length > 0) {
                Type type = array[0];
                if (!(type instanceof Class)) {
                    n3 = 1;
                }
                else if (((Class)type).isInterface()) {
                    n3 = 1;
                }
            }
            Object o;
            for (int i = 0; i < array.length; array2[i] = AnnotatedTypeFactory.buildAnnotatedType(array[i], locationInfo, (TypeAnnotation[])o.toArray(TypeAnnotationParser.EMPTY_TYPE_ANNOTATION_ARRAY), (TypeAnnotation[])fetchBounds.toArray(TypeAnnotationParser.EMPTY_TYPE_ANNOTATION_ARRAY), n), ++i) {
                o = new ArrayList(fetchBounds.size());
                for (TypeAnnotation typeAnnotation : fetchBounds) {
                    TypeAnnotation.TypeAnnotationTargetInfo targetInfo = typeAnnotation.getTargetInfo();
                    if (targetInfo.getSecondaryIndex() == i + n3 && targetInfo.getCount() == n2) {
                        o.add(typeAnnotation);
                    }
                    else {
                        continue;
                    }
                }
            }
            return array2;
        }
        else {
            return new AnnotatedType[0];
        }
    }
    
    private static <D extends GenericDeclaration> List<TypeAnnotation> fetchBounds(D n) {
        TypeAnnotation.TypeAnnotationTarget typeAnnotationTarget;
        Class<Object> clazz;
        if ((n instanceof Class)) {
            typeAnnotationTarget = TypeAnnotation.TypeAnnotationTarget.CLASS_TYPE_PARAMETER_BOUND;
            clazz = (D)n;
        }
        else {
            typeAnnotationTarget = TypeAnnotation.TypeAnnotationTarget.METHOD_TYPE_PARAMETER_BOUND;
            clazz = (D)n;
        }
        return TypeAnnotation.filter(parseAllTypeAnnotations(clazz), typeAnnotationTarget);
    }
    
    static TypeAnnotation[] parseAllTypeAnnotations(AnnotatedElement annotatedElement) {
        JavaLangAccess javaLangAccess = SharedSecrets.getJavaLangAccess();
        Class<Object> declaringClass;
        byte[] array;
        if ((annotatedElement instanceof Class)) {
            declaringClass = (Class)annotatedElement;
            array = javaLangAccess.getRawClassTypeAnnotations(declaringClass);
        }
        else if ((annotatedElement instanceof Executable)) {
            declaringClass = ((Executable)annotatedElement).getDeclaringClass();
            array = javaLangAccess.getRawExecutableTypeAnnotations((Executable)annotatedElement);
        }
        else {
            return TypeAnnotationParser.EMPTY_TYPE_ANNOTATION_ARRAY;
        }
        return parseTypeAnnotations(array, javaLangAccess.getConstantPool(declaringClass), annotatedElement, declaringClass);
    }
    
    private static TypeAnnotation[] parseTypeAnnotations(byte[] array, ConstantPool constantPool, AnnotatedElement annotatedElement, Class<?> clazz) {
        if (array == null) {
            return TypeAnnotationParser.EMPTY_TYPE_ANNOTATION_ARRAY;
        }
        ByteBuffer wrap = ByteBuffer.wrap(array);
        int initialCapacity = wrap.getShort() & 0xFFFF;
        ArrayList list = new ArrayList(initialCapacity);
        for (int i = 0; i < initialCapacity; ++i) {
            TypeAnnotation typeAnnotation = parseTypeAnnotation(wrap, constantPool, annotatedElement, clazz);
            if (typeAnnotation != null) {
                list.add(typeAnnotation);
            }
        }
        return (TypeAnnotation[])list.toArray(TypeAnnotationParser.EMPTY_TYPE_ANNOTATION_ARRAY);
    }
    
    static Map<Class<? extends Annotation>, Annotation> mapTypeAnnotations(TypeAnnotation[] array) {
        LinkedHashMap<Class, Annotation> linkedHashMap = new LinkedHashMap();
        int length = array.length;
        int i = 0;
        while (i < length) {
            Annotation annotation = array[i].getAnnotation();
            Class<Object> annotationType = annotation.annotationType();
            if (AnnotationType.getInstance(annotationType).retention() == RetentionPolicy.RUNTIME && linkedHashMap.put(annotationType, annotation) != null) {
                throw new AnnotationFormatError("Duplicate annotation for class: " + annotationType + ": " + annotation);
            }
            ++i;
        }
        return linkedHashMap;
    }
    
    private static TypeAnnotation parseTypeAnnotation(ByteBuffer byteBuffer, ConstantPool constantPool, AnnotatedElement annotatedElement, Class<?> clazz) {
        try {
            TypeAnnotation.TypeAnnotationTargetInfo targetInfo = parseTargetInfo(byteBuffer);
            TypeAnnotation.LocationInfo locationInfo = TypeAnnotation.LocationInfo.parseLocationInfo(byteBuffer);
            Annotation annotation = AnnotationParser.parseAnnotation(byteBuffer, constantPool, clazz, false);
            if (targetInfo == null) {
                return null;
            }
            return new TypeAnnotation(targetInfo, locationInfo, annotation, annotatedElement);
        }
        catch (IllegalArgumentException | BufferUnderflowException cause) {
            throw new AnnotationFormatError(cause);
        }
    }
    
    private static TypeAnnotation.TypeAnnotationTargetInfo parseTargetInfo(ByteBuffer byteBuffer) {
        int n = byteBuffer.get() & 0xFF;
        switch (n) {
            case 0:
            case 1: {
                int n2 = byteBuffer.get() & 0xFF;
                if (n == 0) {
                    return new TypeAnnotation.TypeAnnotationTargetInfo(TypeAnnotation.TypeAnnotationTarget.CLASS_TYPE_PARAMETER, n2);
                }
                return new TypeAnnotation.TypeAnnotationTargetInfo(TypeAnnotation.TypeAnnotationTarget.METHOD_TYPE_PARAMETER, n2);
            }
            case 16: {
                int short1 = byteBuffer.getShort();
                if (short1 == -1) {
                    return new TypeAnnotation.TypeAnnotationTargetInfo(TypeAnnotation.TypeAnnotationTarget.CLASS_EXTENDS);
                }
                if (short1 >= 0) {
                    return new TypeAnnotation.TypeAnnotationTargetInfo(TypeAnnotation.TypeAnnotationTarget.CLASS_IMPLEMENTS, short1);
                }
                break;
            }
            case 17: {
                return parse2ByteTarget(TypeAnnotation.TypeAnnotationTarget.CLASS_TYPE_PARAMETER_BOUND, byteBuffer);
            }
            case 18: {
                return parse2ByteTarget(TypeAnnotation.TypeAnnotationTarget.METHOD_TYPE_PARAMETER_BOUND, byteBuffer);
            }
            case 19: {
                return new TypeAnnotation.TypeAnnotationTargetInfo(TypeAnnotation.TypeAnnotationTarget.FIELD);
            }
            case 20: {
                return new TypeAnnotation.TypeAnnotationTargetInfo(TypeAnnotation.TypeAnnotationTarget.METHOD_RETURN);
            }
            case 21: {
                return new TypeAnnotation.TypeAnnotationTargetInfo(TypeAnnotation.TypeAnnotationTarget.METHOD_RECEIVER);
            }
            case 22: {
                return new TypeAnnotation.TypeAnnotationTargetInfo(TypeAnnotation.TypeAnnotationTarget.METHOD_FORMAL_PARAMETER, byteBuffer.get() & 0xFF);
            }
            case 23: {
                return parseShortTarget(TypeAnnotation.TypeAnnotationTarget.THROWS, byteBuffer);
            }
            case 64:
            case 65: {
                for (int short2 = byteBuffer.getShort(), i = 0; i < short2; ++i) {
                    byteBuffer.getShort();
                    byteBuffer.getShort();
                    byteBuffer.getShort();
                }
                return null;
            }
            case 66: {
                byteBuffer.get();
                return null;
            }
            case 67:
            case 68:
            case 69:
            case 70: {
                byteBuffer.getShort();
                return null;
            }
            case 71:
            case 72:
            case 73:
            case 74:
            case 75: {
                byteBuffer.getShort();
                byteBuffer.get();
                return null;
            }
        }
        throw new AnnotationFormatError("Could not parse bytes for type annotations");
    }
    
    private static TypeAnnotation.TypeAnnotationTargetInfo parseShortTarget(TypeAnnotation.TypeAnnotationTarget typeAnnotationTarget, ByteBuffer byteBuffer) {
        return new TypeAnnotation.TypeAnnotationTargetInfo(typeAnnotationTarget, byteBuffer.getShort() & 0xFFFF);
    }
    
    private static TypeAnnotation.TypeAnnotationTargetInfo parse2ByteTarget(TypeAnnotation.TypeAnnotationTarget typeAnnotationTarget, ByteBuffer byteBuffer) {
        return new TypeAnnotation.TypeAnnotationTargetInfo(typeAnnotationTarget, byteBuffer.get() & 0xFF, byteBuffer.get() & 0xFF);
    }
    
    static {
        EMPTY_TYPE_ANNOTATION_ARRAY = new TypeAnnotation[0];
    }
}
